// tarpinis kodas, taip pat visu TNode code() funkcijos
// siek tiek opkodu optimizavimo

#include <iostream>
#include <iomanip>

#include "syntree.h"
#include "intcode.h"

using namespace std;

long __locsize;

long TIntInsn::__num = 0;

TIntInsn *TIntInsn::get(long n)
{
		long j = 0;
		TIntInsn *i = this;
		while (i && j != n) {
				i = i->next;
				j++;
		}
		if (!i)
				return NULL;
		return i;
}

// sujungia du instrukciju blokus
TIntInsn *intConcat(TIntInsn *i1, TIntInsn *i2)
{
		TIntInsn *i = i1;
		while (i->next)
				i = i->next;
		i->next = i2;
		return i1;
}

const char *TIntInsn::names[] = {
"NOP", "SAVE", "PUSH", "REFFR", "REFTHIS", "REFSTAT", "REFADD",
"DEL", "GROW", "CALL", "CALLSUPER", "NATIVE", "RET", "NEW", "IS", "DELETE",
// arith, konv ...
"PUSHs", "PUSHif", "PUSHarr", "i2f", "f2i", "JMP", "JZ",
"OR", "XOR", "AND", "SHL", "SHR", "MOD", "ADDi", "SUBi", "MULi", "DIVi",
"BORi", "BANDi", "BEi", "BNEi", "BLi", "BGi", "BLEi", "BGEi",
"ADDf", "SUBf", "MULf", "DIVf",
"BORf", "BANDf", "BEf", "BNEf", "BLf", "BGf", "BLEf", "BGEf",
"NEGi", "BNOTi", "INV",
"SAVEMULi", "SAVEDIVi", "SAVEMOD", "SAVEADDi", "SAVESUBi", // i: += -= ...
"SAVESHL", "SAVESHR", "SAVEOR", "SAVEXOR", "SAVEAND",
"NEGf", "BNOTf",
"SAVEMULf", "SAVEDIVf", "SAVEADDf", "SAVESUBf", // f: +- -= ...
// ++/--
"PREINCi", "PREDECi", "POSINCi", "POSDECi",
"PREINCf", "PREDECf", "POSINCf", "POSDECf",
// advanced
"MULADDf", "MULDIVf", "POP",
"OFSTHIS", "OFSFR", "OFSPTR",
"SAVE1", "PUSH1", "POP1", "DEL1",
// inline
"RAND", "ABS", "SGN",
"MIN", "MAX",
"CLAMP",
"FABS", "FSGN", "SIN", "COS", "TAN", "ATAN", "EXP",
"LOGE", "SQRT", "SQR",
"FRAND",
"FMIN", "FMAX",
"FCLAMP", "LINERP", "COSERP",
"PRINT",
"  label:"
};
const long TIntInsn::isCount[] = {
0, 1, 1, 0, 0, 0, 0,
1, 1, 1, 1, 1, 0, 1, 1, 0,
// arith, konv...
1, 0, 1, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // int 2
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, // flt 2
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, // int 1
0, 0, 0, 0, 0, // i: += -= ...
0, 0, 0, 0, 0,
0, 0, // flt 1
0, 0, 0, 0, // f: +- -= ...
// ++/--
0, 0, 0, 0,
0, 0, 0, 0,
// advanced
0, 0, 1,
0, 0, 0,
0, 0, 0, 0,
// inline
0, 0, 0, 0, 0, 0, // int
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // flt
0,
0
};

void TIntInsn::show()
{
		cout<<"         "<<setw(5)<<num<<' '<<setw(3)<<op<<' '
			<<setw(9)<<setiosflags(ios::left)<<names[op]<<setiosflags(ios::right);
		if (isCount[op])
			cout<<count<<"  ";
		if (op == IOP_PUSHif)
			cout<<vali<<" ("<<(float)valf<<")   ";
		else if (op == IOP_OFSTHIS || op == IOP_OFSFR ||
				 op == IOP_OFSPTR || op == IOP_CALLSUPER)
			cout<<vali;
		else if (target)
			cout<<"__"<<target->num<<"__";
		cout<<endl;
		if (next)
			next->show();
}

void TFunction::code()
{
		// padidinam steka tiek, kiek lokaliu kint.
		TIntInsn *i = NULL;
		__locsize = tab->Size() - tab->Size(args + 3); // args+value+this+addr
		if (__locsize > 0)
				i = new TIntInsn(IOP_GROW, __locsize);
		// sukodinam koda
		if (i)
				op = intConcat(i, tree->code());
		else
				op = tree->code();
		if (isRet(tree)) // kazhkox expr
				intConcat(op, new TIntInsn(IOP_DEL, tree->type()->Size()));
		// del visa ko gale pridedam return. jei turi kazka grazinti, o
		// nera return ???, tai grazina neaisku ka
		// isimam lokalius
		if (__locsize > 0)
				intConcat(op, new TIntInsn(IOP_DEL, __locsize));
		intConcat(op, new TIntInsn(IOP_RET));
}

/*******************
 * VISU NODU KODAS *
 *******************/

TIntInsn *TListNode::code() const
{
		TIntInsn *in = node->code();
		if (isRet(node)) // kazhkox expr
				intConcat(in, new TIntInsn(IOP_DEL, node->type()->Size()));
		intConcat(in, list->code());
		if (isRet(list)) // kazhkox expr
				intConcat(in, new TIntInsn(IOP_DEL, list->type()->Size()));
		return in;
}

TIntInsn *TIfNode::code() const
{
		TIntInsn *chk, *jz, *body, *end;
		chk = cond->code(); 					//	  cond
		jz = new TIntInsn(IOP_JZ);				//	  jz 1
		body = stat->code();					//	  stat
		end = new TIntInsn(JUMPTARGET, jz); 	// 1:
		jz->target = end;
		if (isRet(stat)) // kazhkox expr
				intConcat(body, new TIntInsn(IOP_DEL, stat->type()->Size()));
		intConcat(chk, jz);
		intConcat(jz, body);
		intConcat(body, end);
		return chk;
}
TIntInsn *TIfElseNode::code() const
{
		TIntInsn *chk, *jz, *body, *jmp, *jt, *body2, *end;
		chk = cond->code(); 							//	  cond
		jz = new TIntInsn(IOP_JZ);						//	  jz 1
		body = stat->code();							//	  stat
		jmp = new TIntInsn(IOP_JMP);					//	  jmp 2
		jt = new TIntInsn(JUMPTARGET, jz);				// 1:
		body2 = stat2->code();							//	  stat2
		end = new TIntInsn(JUMPTARGET, jmp);			// 2:
		jz->target = jt;
		jmp->target = end;
		if (isRet(stat)) // kazhkox expr
				intConcat(body, new TIntInsn(IOP_DEL, stat->type()->Size()));
		if (isRet(stat2)) // kazhkox expr
				intConcat(body2, new TIntInsn(IOP_DEL, stat2->type()->Size()));
		intConcat(chk, jz);
		intConcat(jz, body);
		intConcat(body, jmp);
		intConcat(jmp, jt);
		intConcat(jt, body2);
		intConcat(body2, end);
		return chk;
}
TIntInsn *TTernaryNode::code() const
{
		TIntInsn *chk, *jz, *body, *jmp, *jt, *body2, *end;
		chk = cond->code(); 							//	  cond
		jz = new TIntInsn(IOP_JZ);						//	  jz 1
		body = stat->code();							//	  stat
		jmp = new TIntInsn(IOP_JMP);					//	  jmp 2
		jt = new TIntInsn(JUMPTARGET, jz);				// 1:
		body2 = stat2->code();							//	  stat2
		end = new TIntInsn(JUMPTARGET, jmp);			// 2:
		jz->target = jt;
		jmp->target = end;
		intConcat(chk, jz);
		intConcat(jz, body);
		intConcat(body, jmp);
		intConcat(jmp, jt);
		intConcat(jt, body2);
		intConcat(body2, end);
		return chk;
}
TIntInsn *TWhileNode::code() const
{
		TIntInsn *jt, *chk, *jmpz, *body, *end, *jmp;
		jt = new TIntInsn(JUMPTARGET);		  // 0:
		chk = cond->code(); 				  //	cond
		jmpz = new TIntInsn(IOP_JZ);		  //	jz 1
		body = stat->code();				  //	stat
		jmp = new TIntInsn(IOP_JMP, jt);	  //	jmp 0
		end = new TIntInsn(JUMPTARGET, jmpz); // 1:
		jt->target = jmp;
		jmpz->target = end;
		if (isRet(stat)) // kazhkox expr
				intConcat(body, new TIntInsn(IOP_DEL, stat->type()->Size()));
		intConcat(jt, chk);
		intConcat(chk, jmpz);
		intConcat(jmpz, body);
		intConcat(body, jmp);
		intConcat(jmp, end);
		return jt;
}
TIntInsn *TDoNode::code() const
{
		TIntInsn *jt, *body, *chk, *jmpz;
		jt = new TIntInsn(JUMPTARGET);	   // 0:
		body = stat->code();			   //	 stat
		chk = cond->code(); 			   //	 chk
		jmpz = new TIntInsn(IOP_JZ, jt);   //	 jz 0
		jt->target = jmpz;
		if (isRet(stat)) // kazhkox expr
				intConcat(body, new TIntInsn(IOP_DEL, stat->type()->Size()));
		intConcat(jt, body);
		intConcat(body, chk);
		intConcat(chk, jmpz);
		return jt;
}
TIntInsn *TForNode::code() const
{
		TIntInsn *in, *jt, *chk, *jmpz, *body, *lo, *jmp, *end;
		in = init->code();						//	  init
		jt = new TIntInsn(JUMPTARGET);			// 0:
		chk = cond->code(); 					//	  cond
		jmpz = new TIntInsn(IOP_JZ);			//	  jz 1
		body = stat->code();					//	  stat
		lo = loop->code();						//	  loop
		jmp = new TIntInsn(IOP_JMP, jt);		//	  jmp 0
		end = new TIntInsn(JUMPTARGET, jmpz);	// 1:
		intConcat(in, new TIntInsn(IOP_DEL, init->type()->Size()));
		intConcat(lo, new TIntInsn(IOP_DEL, loop->type()->Size()));
		if (isRet(stat)) // kazhkox expr
				intConcat(body, new TIntInsn(IOP_DEL, stat->type()->Size()));
		jt->target = jmp;
		jmpz->target = end;
		intConcat(in, jt);
		intConcat(jt, chk);
		intConcat(chk, jmpz);
		intConcat(jmpz, body);
		intConcat(body, lo);
		intConcat(lo, jmp);
		intConcat(jmp, end);
		return in;
}
TIntInsn *TDeleteNode::code() const
{
		TIntInsn *i1;
		i1 = lval->code();
		return intConcat(i1, new TIntInsn(IOP_DELETE));
}

// BINARINIU KODAS

TIntInsn *TBinaryNode::code() const
{
		TIntInsn *o1 = op1->code();
		TIntInsn *o2 = op2->code();
		TIntInsn *o3 = new TIntInsn((op1->type()==TypeFloat) ? p->opf:p->opi);
		intConcat(o1, o2);
		intConcat(o2, o3);
		return o1;
}

// UNARINIU KODAS

TIntInsn *TUnaryNode::code() const
{
		TIntInsn *o1 = op->code();
		TIntInsn *o2 = new TIntInsn((op->type()==TypeFloat) ? p->opf:p->opi);
		intConcat(o1, o2);
		return o1;
}

// PRISKYRIMU KODAS

TIntInsn *TAssignNode::code() const
{
	TIntInsn *i1, *i2;
	long size = expr->type()->Size();
	if( isMultiExpr( expr ) )
		size = isMultiExpr( expr )->TotalSize();
	i1 = expr->code();
	i2 = lval->code();
	intConcat(i2, new TIntInsn(IOP_SAVE, size));
	return intConcat(i1, i2);
}

TIntInsn *TAsArithNode::code() const
{
		TIntInsn *o1 = expr->code();
		TIntInsn *o2 = lval->code();
		TIntInsn *o3 = new TIntInsn((lval->type()==TypeFloat) ? p->opf:p->opi);
		intConcat(o1, o2);
		intConcat(o2, o3);
		return o1;
}

// KONVERSIJU KODAS

TIntInsn *T2Long::code() const
		{ return intConcat(op->code(), new TIntInsn(IOP_f2i)); }
TIntInsn *T2Float::code() const
		{ return intConcat(op->code(), new TIntInsn(IOP_i2f)); }

// LVALUE KODAS

TIntInsn *TOffsetNode::code() const
{
		TIntInsn *i1, *i2;
		i1 = lval->code();
		i2 = expr->code();
		if (lval->type()->Size() > 1) {
				TIntInsn *i3 = new TIntInsn(IOP_PUSHif, 0, lval->type()->Size());
				intConcat(i3, new TIntInsn(IOP_MULi));
				intConcat(i2, i3);
		}
		intConcat(i2, new TIntInsn(IOP_REFADD));
		return intConcat(i1, i2);
}
TIntInsn *TMemberNode::code() const
{
		TIntInsn *i1, *i2;
		i1 = lval->code();
		i2 = new TIntInsn(IOP_PUSHif, 0, sym->Offset());
		intConcat(i2, new TIntInsn(IOP_REFADD));
		return intConcat(i1, i2);
}
TIntInsn *TIdentNode::code() const
{
	TIntInsn *i1, *i2;
	i1 = lval->code();
	i2 = new TIntInsn(IOP_PUSH, lval->type()->Size());
	return intConcat(i1, i2);
}
TIntInsn *TBaseRefNode::code() const
{
	EIntOpcode o = IOP_REFTHIS;
	switch (mode) {
	case E_THIS   : o = IOP_REFTHIS; break;
	case E_FRAME  : o = IOP_REFFR;	 break;
	case E_STATIC : o = IOP_REFSTAT; break;
	}
	return new TIntInsn(o);
}
TIntInsn *TNewNode::code() const
{
		return new TIntInsn(IOP_NEW, cla->Id());
}
TIntInsn *TIsNode::code() const
{
		return intConcat(lval->code(), new TIntInsn(IOP_IS, cla->Id()));
}

// FUNKCIJU KVIETIMAS

static EIntOpcode builtOps[] = {
		IOP_NOP,
		IOP_RAND, IOP_ABS, IOP_SGN,
		IOP_MIN, IOP_MAX,
		IOP_CLAMP,
		IOP_FABS, IOP_FSGN, IOP_SIN, IOP_COS, IOP_TAN, IOP_ATAN, IOP_EXP,
		IOP_LOGE, IOP_SQRT, IOP_SQR,
		IOP_FRAND,
		IOP_FMIN, IOP_FMAX,
		IOP_FCLAMP, IOP_LINERP, IOP_COSERP,
		IOP_PRINT,
};

TIntInsn *TFuncNode::code() const
{
		long sz = func->Type()->Size(); // graz. reiksmes dydis
		TArgs *a = args;
		TIntInsn *i1, *i2, *i3, *i4, *i5;
		i1 = new TIntInsn(IOP_GROW, sz); // vieta rezultatui
		// darom arg. koda
		sz = 0;
		i2 = i3 = NULL;
		while (a) {
				sz += a->expr->type()->Size();
				i4 = a->expr->code();
				if (!i2)
						i2 = i3 = i4;
				else
						intConcat(i3, i4);
				i3 = i4;
				a = a->next;
		}
		// ptr. i objekta
		i3 = lval->code(); sz++;
		// chia kazhkuri ish builtin inline f-ju
		if (func->Builtin() != EB_NONE) {
				delete i1; // nereikia grow
				i4 = new TIntInsn(builtOps[func->Builtin()]); // opkodas
				if (func->Noref()) // nereikia this
						delete i3;
				else
						i4 = intConcat(i3, i4);
				if (i2)
						i4 = intConcat(i2, i4);
				return i4;
		}
		// call
		if (super < 0)
				i4 = new TIntInsn(IOP_CALL, func->Offset());
		else
				i4 = new TIntInsn(IOP_CALLSUPER, func->Offset(), super);
		i5 = new TIntInsn(IOP_DEL, sz);
		intConcat(i4, i5);
		intConcat(i3, i4);
		if (i2) {
				intConcat(i2, i3);
				intConcat(i1, i2);
		} else
				intConcat(i1, i3);
		return i1;
}

TIntInsn *TNativeNode::code() const
		{ return new TIntInsn(IOP_NATIVE, slot); }

TIntInsn *TReturnNode::code() const
{
		TIntInsn *i;
		if (__locsize > 0) {
				i = new TIntInsn(IOP_DEL, __locsize);
				intConcat(i, new TIntInsn(IOP_RET));
		} else
				i = new TIntInsn(IOP_RET);
		return i;
}

// MULTI-EXPR

TIntInsn *TMultiExprNode::code() const
{
	TArgs *a = exprs;
	TIntInsn *i, *i3, *i4;
	// darom expr. koda
	i = i3 = NULL;
	while( a ) {
		i4 = a->expr->code();
		if (!i)
			i = i3 = i4;
		else
			intConcat(i3, i4);
		i3 = i4;
		a = a->next;
	}
	return i;
}

// TERMINATORIAI

TIntInsn *TNode::code() const // dummy - nop
		{ return new TIntInsn; };

TIntInsn *TStringNode::code() const
		{ return new TIntInsn(IOP_PUSHs, offset); };
TIntInsn *TNumberNode::code() const
		{ return new TIntInsn(IOP_PUSHif, 0, (long)val); };
TIntInsn *TFloatNode::code() const
		{ return new TIntInsn(IOP_PUSHif, val); };

/*******************************
 * TARPINIO KODO OPTIMIZAVIMAS *
 *******************************/

TIntInsn *TIntInsn::optimize()
{
		TIntInsn *i;
		if (!next)
				return this;
		do {
				i = next;
				next = next->optimize();
		} while (next != i);
		i = next->next;
		// optimizuojam
		long tmp1, tmp2;
		switch (op) {
		// nuimam NOP
		case IOP_NOP:
				i = next;
				delete this;
				return i;
		// grow + grow | grow + del | grow 0
		// grow 1 + del1 => ...
		case IOP_GROW:
				if (next->op == IOP_GROW) {
						count += next->count;
						delete next;
						next = i;
				} else if (next->op == IOP_DEL) {
						count -= next->count;
						if (count < 0) {
								count = -count;
								op = IOP_DEL;
						}
						delete next;
						next = i;
				} else if (count == 1 && next->op == IOP_DEL1) {
						i = next->next;
						delete next;
						delete this;
						return i;
				}
				if (!count) {
						i = next;
						delete this;
						return i;
				}
				return this;
		// del + del | del + grow | del 0
		// del1 => del1
		case IOP_DEL:
				if (next->op == IOP_DEL) {
						count += next->count;
						delete next;
						next = i;
				} else if (next->op == IOP_GROW) {
						count -= next->count;
						if (count < 0) {
								count = -count;
								op = IOP_GROW;
						}
						delete next;
						next = i;
				}
				if (!count) {
						i = next;
						delete this;
						return i;
				}
				if (count == 1 && op == IOP_DEL) op = IOP_DEL1;
				return this;
		// ret
		case IOP_RET:
				// visas iki jumpu nuimam
				TIntInsn *i2;
				i = next;
				while (i && !i->target) {
						i2 = i->next;
						delete i;
						i = i2;
				}
				next = i;
				return this;
		// refthis + pushi ? + refadd => ofsthis ?
		// reffr + pushi ? + refadd => ofsfr ?
		case IOP_REFTHIS:
		case IOP_REFFR:
				if (get(2) && get(1)->op == IOP_PUSHif && get(2)->op == IOP_REFADD) {
						i = get(2);
						i->op	= (op==IOP_REFTHIS) ? IOP_OFSTHIS : IOP_OFSFR;
						i->vali = get(1)->vali;
						delete get(1);
						delete this;
						return i;
				}
				return this;
		// push 1 + pushi ? + refadd => ofsptr ?
		// push 1 => push1
		case IOP_PUSH:
				if (count == 1 && get(2) && get(1)->op == IOP_PUSHif &&
											get(2)->op == IOP_REFADD) {
						i = get(2);
						i->op	= IOP_OFSPTR;
						i->vali = get(1)->vali;
						delete get(1);
						delete this;
						return i;
				}
				if (count == 1) op = IOP_PUSH1;
				return this;
		// pushi 1 mul/div =>
		// pushi mul/div   => pushi shl/shr
		case IOP_PUSHif:
				if (next->op == IOP_MULi || next->op == IOP_DIVi) {
						if (vali == 1) {
								i = next->next;
								delete next;
								delete this;
								return i;
						}
						tmp1 = vali;
						tmp2 = 0;
						while (!(tmp1 & 1)) {
								tmp1 >>= 1;
								tmp2++;
						}
						if (tmp1 == 1) { // tinka
								vali = tmp2;
								if (next->op == IOP_MULi)
										next->op = IOP_SHL;
								else
										next->op = IOP_SHR;
						}
				}
				return this;
		// mulf addf => muladdf
		// mulf divf => muldivf
		case IOP_MULf:
				if (next->op == IOP_ADDf) {
						delete next;
						next = i;
						op = IOP_MULADDf;
				} else if (next->op == IOP_DIVf) {
						delete next;
						next = i;
						op = IOP_MULDIVf;
				}
				return this;
		// save del => pop
		// save 1 => save1
		case IOP_SAVE:
				if (next->op == IOP_DEL && count == next->count) {
						delete next;
						next = i;
						op = IOP_POP;
				}
				if (count == 1) op = IOP_SAVE1;
				return this;
		// save1 del1 => pop1
		case IOP_SAVE1:
				if (next->op == IOP_DEL1) {
						delete next;
						next = i;
						op = IOP_POP1;
				}
				return this;
		// pop 1 => pop1
		case IOP_POP:
				if (count == 1) op = IOP_POP1;
				return this;
		default:
				return this;
		}
		return this;
}

